# SPDX-FileCopyrightText: 2026 Teilo Millet
# SPDX-License-Identifier: Apache-2.0

"""Battle tests — property-based exploration of the vauban state space.

Declare invariants, explore the state space, inject faults, replay
deterministically.  Run with ``pytest -m battle`` or
``pytest tests/battle/``.
"""
